'// Reverse DNS Scan
'// Created in 2007 by Casey Sprangel using PowerBasic/CC (www.powerbasic.com)
'// A tool to look up a range of IP addresses and return the reverse DNS (if any)
'//----------------------------------------------------------------------------

#COMPILE EXE
#DIM ALL
#INCLUDE "WSOCK32.INC"                                        '// windows api for internet stuff

%myversion = 20090223                                         '// last time this program was updated
$title = "RDNS "                                              '// name of this program including trailing space

MACRO ipdot = FORMAT$(@p[3], "#") + "." + FORMAT$(@p[2], "#") + "." + FORMAT$(@p[1], "#") + "." + FORMAT$(@p, "#")
MACRO ipclass = FORMAT$(@p[3], "#") + "." + FORMAT$(@p[2], "#") + "." + FORMAT$(@p[1], "#")

DECLARE FUNCTION IP_to_Host(haddr AS ASCIIZ) AS STRING
DECLARE FUNCTION isip(b AS STRING) AS LONG
DECLARE FUNCTION getip(b AS STRING, m AS LONG) AS DWORD

'//----------------------------------------------------------------------------

FUNCTION IP_to_Host(haddr AS ASCIIZ) AS STRING                '// IP to Host Name
  LOCAL lpaddr AS in_addr
  LOCAL HostEnt AS HOSTENTSTRU PTR
  LOCAL g_result AS LONG
  LOCAL buf AS STRING

  lpaddr.s_addr = Inet_Addr(haddr)                            '// ascii IP to dword addr
  hostent = GetHostByAddr(lpaddr.s_addr, 4, %PF_INET)         '// get host by address
  IF hostent = 0 THEN                                         '// null =  error
    g_Result = WSAGetLastError                                '// get the error number from winsock
    SELECT CASE g_result
      CASE %WSANOTINITIALISED
        buf = "A successful WSAStartup must occur before using this function.
      CASE %WSAENETDOWN
        buf = "The network subsystem has failed.
      CASE %WSAHOST_NOT_FOUND
        buf = "Authoritative Answer Host not found.
      CASE %WSATRY_AGAIN
        buf = "Non-Authoritative Host not found, or server failed.
      CASE %WSANO_RECOVERY
        buf = "Nonrecoverable error occurred.
      CASE %WSANO_DATA
        EXIT FUNCTION                                         '// leave - this means they have no DNS
        buf = "Valid name, no data record of requested type.
      CASE %WSAEINPROGRESS
        buf = "A blocking Windows Sockets 1.1 call is in progress, or the service provider is still processing a callback function.
      CASE %WSAEAFNOSUPPORT
        buf = "The type specified is not supported by the Windows Sockets implementation.
      CASE %WSAEFAULT
        buf = "The addr argument is not a valid part of the user address space, or the len argument is too small.
      CASE %WSAEINTR
        buf = "The (blocking) call was canceled through WSACancelBlockingCall.
       CASE ELSE
         buf = "Unknown error?"
    END SELECT
    FUNCTION = "Error: " + FORMAT$(g_result) + $SPC + buf
  ELSE
    FUNCTION = @hostent.@h_name
  END IF
END FUNCTION

'//----------------------------------------------------------------------------

FUNCTION isip(b AS STRING) AS LONG                            '// returns true if this is a valid IP address with optional bit mask
  LOCAL a AS STRING
  LOCAL i AS LONG
  LOCAL p1 AS LONG
  LOCAL p2 AS LONG
  LOCAL p3 AS LONG
  LOCAL p4 AS LONG
  LOCAL p5 AS LONG
  IF b = "" THEN                                              '// empty IP address?
    EXIT FUNCTION                                             '// leave, nothing to do
  END IF
  a = TRIM$(b)                                                '// make a copy so we can modify
  i = INSTR(a,"/")                                            '// is there a net mask?
  IF i > 0 THEN
    p5 = VAL(MID$(a, i + 1))                                  '// get the mask
    a = LEFT$(a, i -1)                                        '// remove the extra part
  END IF
  IF TALLY(a, ".") = 3 THEN                                   '// valid number of elements?
    i = INSTR(a, ".")                                         '// locate next period
    p1 = VAL(LEFT$(a, i - 1))                                 '// first number
    a = MID$(a, i + 1)                                        '// trim off digit
    i = INSTR(a, ".")                                         '// locate next period
    p2 = VAL(LEFT$(a, i - 1))                                 '// second number
    a = MID$(a, i + 1)                                        '// trim off digit
    i = INSTR(a, ".")                                         '// locate next period
    p3 = VAL(LEFT$(a, i - 1))                                 '// third number
    a = MID$(a, i + 1)                                        '// trim off digit
    p4 = VAL(a)                                               '// the left over bit is the last digit
    IF p1 > -1 AND p1 < 256 AND _                             '// all of the digits must be between 0 and 255 inclusive
       p2 > -1 AND p2 < 256 AND _
       p3 > -1 AND p3 < 256 AND _
       p4 > -1 AND p4 < 256 AND _
       p5 > -1 AND p5 < 33 THEN                               '// the next mask must be between 0 and 32 inclusive
      FUNCTION = -1                                           '// it checks out ok
    END IF
  END IF
END FUNCTION

'//----------------------------------------------------------------------------

FUNCTION getip(b AS STRING, m AS LONG) AS DWORD               '// returns the requested 32 bit number of the IP address string
  '// b is the ip address in "#.#.#.#/#" format
  '// m is the number to return (1=first, 2=last, 3=the exact IP as given)
  LOCAL a AS STRING
  LOCAL buf AS STRING
  LOCAL i AS LONG
  LOCAL p1 AS LONG                                            '// 1st octet
  LOCAL p2 AS LONG                                            '// 2nd octet
  LOCAL p3 AS LONG                                            '// 3rd octet
  LOCAL p4 AS LONG                                            '// 4th octet
  LOCAL p5 AS LONG                                            '// bits (netmask)
  IF b = "" THEN                                              '// empty IP address?
    EXIT FUNCTION                                             '// leave, nothing to do
  END IF
  a = TRIM$(b)                                                '// make a copy so we can modify
  i = INSTR(a,"/")                                            '// is there a net mask?
  IF i > 0 THEN
    p5 = VAL(MID$(a, i + 1))                                  '// get the number of bits in netmask
    a = LEFT$(a, i -1)                                        '// remove the mask part
  ELSE                                                        '// no netmask?
    p5 = 32                                                   '// force to single address
  END IF
  IF TALLY(a, ".") = 3 THEN                                   '// are there the correct number of parts?
    i = INSTR(a, ".")                                         '// locate next period
    p1 = VAL(LEFT$(a, i - 1))                                 '// first number
    a = MID$(a, i + 1)                                        '// trim off digit
    i = INSTR(a, ".")                                         '// locate next period
    p2 = VAL(LEFT$(a, i - 1))                                 '// second number
    a = MID$(a, i + 1)                                        '// trim off digit
    i = INSTR(a, ".")                                         '// locate next period
    p3 = VAL(LEFT$(a, i - 1))                                 '// third number
    a = MID$(a, i + 1)                                        '// trim off digit
    p4 = VAL(a)                                               '// the left over bit is the last digit
    buf = BIN$(p1,8) + BIN$(p2,8) + BIN$(p3,8) + BIN$(p4,8)   '// build binary string of our ip address
    SELECT CASE m
      CASE 1                                                  '// return first IP address
        FUNCTION = VAL("&b" + LEFT$(buf, p5) + STRING$(32 - p5, "0"))
      CASE 2                                                  '// return last IP address
        FUNCTION = VAL("&b" + LEFT$(buf, p5) + STRING$(32 - p5, "1"))
      CASE ELSE                                               '// everything else
        FUNCTION = VAL("&b" + buf)                            '// return the IP given to us
    END SELECT
  END IF                                                      '// are there the correct number of parts
END FUNCTION

'//----------------------------------------------------------------------------

FUNCTION PBMAIN () AS LONG                                    '// program starts here
  DIM buf AS STRING
  DIM buf2 AS STRING
  DIM hostname AS STRING
  DIM h AS LONG                                               '// file handle
  DIM i AS DWORD
  DIM i2 AS BYTE PTR
  DIM ip1 AS DWORD                                            '// starting ip address
  DIM ip2 AS DWORD                                            '// ending ip address
  DIM p AS BYTE PTR
  DIM showall AS LONG                                         '// flag for verbose mode
  DIM result AS LONG
  DIM tWsaData AS WSADATA
  DIM tHostEnt AS HOSTENTSTRU PTR
  DIM tIn_addr AS in_addr
  DIM failcount AS LONG
  DIM failmax AS LONG
  DIM classtrack AS STRING                                    '// tracks the class C address
  DIM restart AS LONG                                         '// flag to record and use the restart file
'  dim restart_file as string

'  restart_file = trim$($title) + ".RST"                       '// build restart file name
  buf = COMMAND$                                              '// get command line from user
  IF INSTR(buf, "?") > 0 OR buf = "" THEN                     '// did user ask for help or not provide enough info?
    STDOUT $title + "version " + FORMAT$(%myversion)          '// show title and version
    STDOUT "This program looks up the reverse DNS name for a single IP address or a range"
    STDOUT "of IP addresses. The output is sent to the screen, but may be redirected to"
    STDOUT "a file if desired."
    STDOUT ""
    STDOUT "Syntax: RDNS <ip_address>[/<bits>] [/F#] [/R] [/V]"
    STDOUT ""
    STDOUT "The IP address is any valid dotted 32 bit address you want to lookup. If"
    STDOUT "the optional bits value is given, all of the IP addresses in the range are"
    STDOUT "resolved in sequence."
    STDOUT ""
    STDOUT "The failure /F switch sets the number of failed lookups to allow before"
    STDOUT "skipping to the next class C block of addresses. Valid range is 0 (default)"
    STDOUT "to 255."
    STDOUT ""
    STDOUT "The restart /R switch starts the scanning at the IP address given on the"
    STDOUT "command line instead of the first host address in the block defined by"
    STDOUT "the BITS option."
    STDOUT ""
    STDOUT "The verbose /V switch shows all IP addresses, even those without a reverse"
    STDOUT "DNS entry."
    EXIT FUNCTION                                             '// leave program
  END IF
  PRINT $title + "version " + FORMAT$(%myversion)             '// this may not be redirected
  i = INSTR(UCASE$(buf), "/F")                                '// locate failure switch
  IF i > 0 THEN                                               '// found switch
    IF MID$(buf, i + 2) = $SPC THEN
      STDOUT "Syntax Error: A number must immediately follow the /F parameter."
      FUNCTION = 1                                            '// set errorlevel 1 on exit
      EXIT FUNCTION                                           '// leave program
    END IF
    failmax = ABS(INT(VAL(MID$(buf, i + 2))))                 '// get counter value
    IF failmax < 1 OR failmax > 255 THEN                      '// is counter outside accepted range?
      failmax = 0                                             '// set default to zero (no fail)
    END IF
    ip1 = INSTR(i, buf + $SPC, ANY "/" + $SPC)                '// find end of this parameter, next parameter or end of string
    buf = TRIM$(LEFT$(buf, i - 1) + MID$(buf, ip1))           '// cut out the switch
  END IF                                                      '// found verbose switch
  i = INSTR(UCASE$(buf), "/V")                                '// locate verbose switch
  IF i > 0 THEN                                               '// found switch
    showall = -1                                              '// set flag to true
    buf = TRIM$(LEFT$(buf, i - 1) + MID$(buf, i + 2))         '// cut out the switch
  END IF                                                      '// found verbose switch
  i = INSTR(UCASE$(buf), "/R")                                '// locate restart switch
  IF i > 0 THEN                                               '// found switch
    restart = -1                                              '// set flag to true
    buf = TRIM$(LEFT$(buf, i - 1) + MID$(buf, i + 2))         '// cut out the switch
  END IF                                                      '// found verbose switch
  IF NOT isip(buf) THEN                                       '// not a valid ip address + netmask?
    STDOUT "Error: Invalid command line syntax. Please use /? for help."
    FUNCTION = 1                                              '// set errorlevel one on exit
    EXIT FUNCTION                                             '// leave program
  END IF                                                      '// not a valid ip address + netmask
  ip1 = getip(buf, 1)                                         '// get starting IP address
  ip2 = getip(buf, 2)                                         '// get ending IP address (returns the first one if no mask is given)
  IF restart THEN                                             '// are we in restart mode?
'    IF ISFILE(restart_file) THEN                              '// is restart file here?
'      h = FREEFILE                                            '// get next available file handle
'      OPEN restart_file FOR INPUT AS h                        '// open the file
'      INPUT #h, i                                             '// get the last used IP
'      CLOSE h                                                 '// release the file handle
'      IF i > ip1 THEN                                         '// use this only if it is more than the default starting IP
'        ip1 = i
'      ELSE
'        STDOUT "Warning: The information in " + restart_file + " appears to be corrupted and was ignored."
'      END IF
'    ELSE                                                      '// restart file does not exist
      ip1 = getip(buf, 3)                                     '// get the starting IP address given by the user
'    END IF
  END IF
  p = VARPTR(ip1)                                             '// get pointer to start
  PRINT "Scanning " + ipdot;
  IF ip2 > ip1 THEN                                           '// is ending IP address different?
    p = VARPTR(ip2)                                           '// get pointer to end
    PRINT " to " + ipdot;
  END IF                                                      '// is ending IP address different
  PRINT " (Press Esc to abort.)"
  result = WSAStartup(257, tWsaData)                          '// start winsock
  IF result <> 0 THEN                                         '// was there an error?
    STDOUT "Error: WSAStartup failed. Aborting."
    FUNCTION = 1                                              '// set errorlevel 1
    EXIT FUNCTION                                             '// leave
  END IF
  p = VARPTR(i)                                               '// get pointer to counter for quick number/string conversion
  FOR i = ip1 TO ip2                                          '// loop through all of our IP addresses
    IF failcount > failmax THEN                               '// too many failed lookups on this class C block?
      IF showall THEN                                         '// does user want to see extra processing information?
        STDOUT ": Too many failures. Skipping remainder of this class C block."
      END IF
      DO WHILE classtrack = ipclass                           '// skip to the next class C block
        INCR i                                                '// move to next IP address 32 bit value
      LOOP
    END IF
'    IF restart THEN                                           '// are we in restart mode?
'      h = FREEFILE                                            '// get next available file handle
'      OPEN restart_file FOR output AS h                       '// open the file
'      ?#h, i                                                  '// save the IP
'      CLOSE h                                                 '// release the file handle
'    END IF
    buf = ipdot                                               '// turn 32-bit number into dotted IP address
    IF classtrack <> ipclass THEN                             '// did the class change?
      classtrack = ipclass                                    '// remember new class
      failcount = 0                                           '// reset failed lookup counter
    END IF
    CONSOLE NAME $title + buf                                 '// show progress on the title bar
    hostname = ip_to_host(buf + $NUL)                         '// lookup reverse DNS for this IP address
    IF hostname > "" AND LEFT$(hostname,6) <> "Error:" THEN
      STDOUT buf + $SPC + hostname                            '// show IP address and hostname
      failcount = 0                                           '// forget any failures and start counting over
    ELSE
      IF failmax > 0 THEN                                     '// do we care about counting errors?
        INCR failcount                                        '// count the number of failed lookups we have
      END IF
      IF showall THEN                                         '// do we have anything to show?
        STDOUT buf + $SPC + $DQ + hostname + $DQ              '// show IP address and hostname
      END IF
    END IF
    buf = INKEY$                                              '// look for a key press on the keyboard
    IF buf = $ESC THEN                                        '// did user press the escape key?
      PRINT "* User abort with Escape key *"
      EXIT FOR                                                '// leave loop
    END IF
  NEXT
  CALL WsaCleanup()                                           '// close winsock
END FUNCTION
